from __future__ import annotations

import uuid
from pathlib import Path
from typing import TypeVar
from urllib.parse import urlparse

T = TypeVar("T")


def is_url(path):
    prefixes = ["http://", "https://", "www."]
    return any(path.startswith(prefix) for prefix in prefixes)


def detect_passing_local_file_as_str(input: T) -> T | Path:
    if isinstance(input, str):
        try:
            parsed_url = urlparse(input)
            if parsed_url.scheme:
                # Do not do any rewriting if an explicit URL is passed in
                return input
        except:  # noqa: E722
            # Fall through to checking whether it exists as file
            pass

        as_path = Path(input)
        is_maybe_file = as_path.exists()

        if is_maybe_file:
            return as_path
    return input


def append_url_part(base: str, part: str) -> str:
    if base.endswith("/"):
        return f"{base[:-1]}/{part}"
    else:
        return f"{base}/{part}"


def is_uuid(s: str):
    try:
        uuid.UUID(s)
    except (ValueError, TypeError):
        return False
    return True



import json
import math
import sys
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Literal, Union

import fused._global_api
from fused._optional_deps import HAS_PANDAS
from fused.api import FusedAPI
from fused.models.udf.udf import Udf


def invoke_lite(api: FusedAPI, function_name, payload, seq):
    t0 = time.perf_counter()

    try:
        response = api._run_lite(function_name, payload)
        dur = time.perf_counter() - t0
        return response, dur, seq
    except Exception as exc:
        dur = time.perf_counter() - t0
        return [f"ERR: {str(exc)}"] * len(payload), dur, seq


def lite_batch(
    url_or_udf: Union[str, Path, Udf],
    batch_size: int,
    arg_list: list[dict],
    processor: Literal["duckdb", "pandas"] = "pandas",
):
    """Main function to run the batch invocation.

    Args:
        url_or_udf: A UDF identifier, name, token, url or UDF object. If not a UDF object use `fused.load` to fetch it.
        batch_size: Payload size sent to each invocation
        arg_list: A list of input parameters for the UDF specified as a list of dictionaries.
        processor: Main executor to run the UDF

    Example:
        >>> import fused
        >>> from fused._lite import lite_batch
        >>> @fused.udf
        >>> def udf(x):
        >>>     import time
        >>>     time.sleep(0.5)
        >>>     return x
        >>> lite_batch(code, 20, [{"x": i} for i in range(200)])


    """
    invocations = math.ceil(len(arg_list) / batch_size)
    max_workers = invocations
    item_count = len(arg_list)
    processor_name = f"fused-lite-poc-{processor}"

    try:
        if isinstance(url_or_udf, Udf):
            udf = url_or_udf
        else:
            udf = fused.load(url_or_udf)

        lines = udf.code.split("\n")
        # TODO: Possibly install fused on the processor
        filtered_lines = [
            line for line in lines if not line.lstrip().startswith("@fused.udf")
        ]

        code = "\n".join(filtered_lines)

        payload_list = []
        # Add python code to be sent to processor
        # TODO: Update request format
        for i, d in enumerate(arg_list):
            if not isinstance(d, dict):
                raise TypeError(
                    f"Error: Payload item {i + 1} must be a dict, got {type(d).__name__}"
                )
            d["python_code"] = code

            # Create nested structure for each payload
            # TODO: udf_name is currently unused in nested structure
            nested_payload = {"udf_name": "", "udf_kwargs": d}
            payload_list.append(nested_payload)

        batch_start = 0
        batch_payloads = []
        for batch_num in range(1, invocations + 1):
            batch_payload = payload_list[batch_start : batch_start + batch_size]

            # udf_name in outer structure is used to indicate processor
            # TODO: Update request format
            payload_data = {
                "udf_name": processor_name,
                "udf_kwargs": {"payloads": batch_payload},
            }
            batch_payloads.append(payload_data)
            batch_start += batch_size

    except json.JSONDecodeError as e:
        raise ValueError(f"Error: Invalid JSON in payload: {e}")
    except Exception as e:
        raise ValueError(f"Error processing payloads: {e}")

    # Execute all invocations concurrently
    results = []
    api = fused._global_api.get_api()
    with ThreadPoolExecutor(max_workers=max_workers) as exe:
        start = time.perf_counter()

        fut_to_seq = {
            exe.submit(
                invoke_lite, api, "fused-lite-poc-async", batch_payloads[seq - 1], seq
            ): seq
            for seq in range(1, max_workers + 1)
        }

        for fut in as_completed(fut_to_seq):
            results.append(fut.result())

    total_time = time.perf_counter() - start

    successful = 0
    parsed_results = []
    notified_error = False
    for batch_result_list, dur, seq in results:
        for r in batch_result_list:
            if isinstance(r, dict) and r["success"]:
                parsed_result = r["response_payload"]
                successful += 1
            elif isinstance(r, str):
                parsed_result = {"success": False, "body": r}
            else:
                parsed_result = {"success": False, "body": r["error"]}

            parsed_result["batch"] = seq
            parsed_result["duration"] = dur
            parsed_result["function_name"] = processor_name

            parsed_results.append(parsed_result)

            if not notified_error and not parsed_result["success"]:
                print("An error occurred in one or more results", file=sys.stderr)
                notified_error = True

    parsed_results.sort(key=lambda r: (r["success"], r["batch"]))
    failed = len(payload_list) - successful
    throughput = item_count / total_time

    print(f"{successful}/{item_count} succeeded, {failed} failed")
    print(f"Batch size: {batch_size} per worker")
    print(f"Workers: {max_workers}")
    print(f"Total time: {total_time:.2f}s")
    print(f"Throughput: {throughput:.1f} invocations/second")

    return create_results_dataframe(parsed_results)


def create_results_dataframe(results):
    """Create a pandas DataFrame from results."""
    # TODO: Provide alternatives if pandas not available
    if not HAS_PANDAS:
        raise ImportError("pandas is required to use this method")

    import pandas as pd

    if not results:
        return pd.DataFrame()

    df_data = []
    for result in results:
        df_data.append(
            {
                "batch_num": result["batch"],
                "duration_seconds": result["duration"],
                "response": result["body"],
                "success": result["success"],
                "function_name": result["function_name"],
            }
        )

    results_df = pd.DataFrame(df_data)

    # Add some derived columns
    if len(results_df) > 0:
        results_df["duration_ms"] = results_df["duration_seconds"] * 1000
        results_df["rank_by_duration"] = results_df["duration_seconds"].rank(
            ascending=False
        )

    return results_df



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import requests
from loguru import logger

from fused._options import StorageStr, _get_data_dir
from fused._options import options as OPTIONS
from fused.core._impl._context_impl import context_in_batch, context_in_realtime

if TYPE_CHECKING:
    import fsspec


def data_path(storage: StorageStr = "auto") -> Path:
    if storage != "auto":
        return _get_data_dir(storage)
    return OPTIONS.data_directory


def filesystem(protocol: str, **storage_options) -> fsspec.AbstractFileSystem:
    """Get an fsspec filesystem for the given protocol.

    Args:
        protocol: Protocol part of the URL, such as "s3" or "gs".
        storage_options: Additional arguments to pass to the storage backend.

    Returns:
        An fsspec AbstractFileSystem.
    """

    import fsspec

    return fsspec.filesystem(protocol, **storage_options)


def _download_requests(url: str) -> bytes:
    # this function is shared
    response = requests.get(url, headers={"User-Agent": ""})
    response.raise_for_status()
    return response.content


def _download_signed(url: str) -> bytes:
    from fused.api._public_api import get_api

    api = get_api()
    return _download_requests(api.sign_url(url))


def _download_object(protocol: str, url: str) -> bytes:
    """

    Args:
        protocol: Protocol part of the URL, such as "s3" or "gs".
        url: Object URL with or without the protocol.

    Returns:
        The object's content in bytes
    """
    # Local needs to use signed URL to impersonal remote IAM role to download the file while remote can assume it has
    # direct access to S3 resources due to its IAM role.
    if not context_in_realtime() and not context_in_batch():
        logger.debug("Trying a signed URL")
        try:
            return _download_signed(url)
        except Exception as e:
            logger.debug(str(e))

    fs = filesystem(protocol)
    with fs.open(url, "rb") as f:
        return f.read()



import asyncio
import concurrent.futures
import json
import time
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from datetime import datetime, timedelta
from math import floor
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    overload,
)

from requests.exceptions import ReadTimeout

from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._options import options as OPTIONS
from fused._run import ResultType
from fused._run import run as fused_run
from fused._run import run_async as fused_run_async
from fused._udf import udf as fused_udf
from fused._udf.decorators import _parse_ttl
from fused.core._cache import DEFAULT_CACHE_MAX_AGE
from fused.core._cache import cache as fused_cache
from fused.models.udf import Udf
from fused.models.udf.base_udf import BaseUdf
from fused.models.udf.udf import AnyBaseUdf
from fused.types import UdfRuntimeError, UdfSerializationError, UdfTimeoutError
from fused.warnings import (
    FusedDeprecationWarning,
)

if TYPE_CHECKING:
    import pandas as pd

# Check for nest_asyncio availability
# Required for environments with an existing event loop, like jupyter notebooks
# Otherwise calling a sync function that creating async tasks within this environment will deadlock processing these
# tasks since the event loop is blocked by that sync function.
try:
    import nest_asyncio

    HAS_NEST_ASYNCIO = True
except ImportError:
    HAS_NEST_ASYNCIO = False

Status = Literal["cancelled", "running", "timeout", "error", "success", "pending"]
STATUS_NOT_FINISHED = ("running", "pending")
STATUS_FAILED = ("error", "timeout")
ExecutionType = Literal["thread_pool", "async_loop"]


def _coerce_hashable(df: "pd.DataFrame"):
    """Make the values in the DataFrame hashable.
    E.g. Python `list` cannot be hashed."""

    for index in df.index:
        for column in df.columns:
            existing_value = df.loc[index, column]

            if isinstance(existing_value, list):
                # list we know a way to make it a hashable type
                existing_value = tuple(existing_value)

            try:
                # Some other types like `dict` we don't have a strategy
                # for, so instead try coercing to json.
                hash(existing_value)
            except TypeError:
                existing_value = json.dumps(existing_value)

            df.at[index, column] = existing_value

    return df


class BaseFuture(ABC):
    """Base class for Future implementations"""

    def __init__(self, index, args):
        self._index = index
        self._args = args
        self._started_at: Optional[datetime] = None
        self._ended_at: Optional[datetime] = None

    def time(self) -> Optional[timedelta]:
        """How long this future took to complete, or None
        if the future is not complete."""
        if self._started_at is None or self._ended_at is None:
            return None
        return self._ended_at - self._started_at

    @abstractmethod
    def done(self) -> bool:
        pass

    @abstractmethod
    def result(self):
        pass

    @abstractmethod
    def exception(self) -> Optional[Exception]:
        pass

    @abstractmethod
    def logs(self) -> str:
        pass

    @abstractmethod
    def status(self) -> Status:
        pass

    def __repr__(self) -> str:
        return f"<fused.Future [status: {('done - ' if self.done() else '') + self.status()}]>"


class Future(BaseFuture):
    """Thread-based Future implementation"""

    def __init__(self, future, index, args):
        super().__init__(index, args)
        self._future: concurrent.futures.Future = future

    def done(self) -> bool:
        return self._future.done()

    def result(self):
        response = self._future.result()
        if isinstance(response, Exception):
            raise response
        if hasattr(response, "error_message") and response.error_message is not None:
            exc = self.exception()
            if exc is not None:
                raise exc
        return response.data

    def exception(self) -> Optional[Exception]:
        try:
            response = self._future.result()
        except concurrent.futures.CancelledError as e:
            return e

        if isinstance(response, Exception):
            return response
        if hasattr(response, "error_message") and response.error_message is not None:
            if response.error_type == "timeout_error":
                return UdfTimeoutError(response.error_message)
            elif response.error_type == "serialization_error":
                return UdfSerializationError(response.error_message)
            return UdfRuntimeError(
                f"[Run #{self._index} {self._args}] {response.error_message}",
                child_exception_class=getattr(response, "exception_class", None),
            )
        future_exc = self._future.exception()
        return future_exc if isinstance(future_exc, Exception) else None

    def logs(self) -> str:
        response = self._future.result()
        if isinstance(response, Exception):
            return str(response)
        out = ""
        if response.stdout:
            out += "stdout\n------\n" + response.stdout
        if response.stderr:
            out += "\nstderr\n------\n" + response.stderr
        return out

    def status(self) -> Status:
        if self._future.cancelled():
            return "cancelled"
        elif self._future.running():
            return "running"
        elif self._future.done():
            exc = self.exception()
            if exc:
                if isinstance(
                    exc, (ReadTimeout, UdfTimeoutError)
                ) or "504 Server Error: Gateway Timeout" in str(exc):
                    return "timeout"
                if isinstance(exc, concurrent.futures.CancelledError):
                    return "cancelled"
                return "error"
            return "success"  # or "finished"?
        else:
            return "pending"


class AsyncFuture(BaseFuture):
    """Async-based Future implementation"""

    def __init__(self, task, index, args):
        super().__init__(index, args)
        self._task: asyncio.Task = task

    def done(self) -> bool:
        return self._task.done()

    def result(self):
        if not self._task.done():
            raise RuntimeError("Task is not done yet")

        response = self._task.result()
        if isinstance(response, Exception):
            raise response

        if hasattr(response, "error_message") and response.error_message is not None:
            exc = self.exception()
            if exc is not None:
                raise exc
        return response.data

    def exception(self) -> Optional[Exception]:
        if not self._task.done():
            return None

        try:
            response = self._task.result()
        except asyncio.CancelledError as e:
            return e

        if isinstance(response, Exception):
            return response
        if hasattr(response, "error_message") and response.error_message is not None:
            return UdfRuntimeError(
                f"[Run #{self._index} {self._args}] {response.error_message}",
                child_exception_class=getattr(response, "exception_class", None),
            )
        task_exc = self._task.exception()
        return task_exc if isinstance(task_exc, Exception) else None

    def logs(self) -> str:
        if not self._task.done():
            return ""

        response = self._task.result()
        if isinstance(response, Exception):
            return str(response)

        out = ""
        if response.stdout:
            out += "stdout\n------\n" + response.stdout
        if response.stderr:
            out += "\nstderr\n------\n" + response.stderr
        return out

    def status(self) -> Status:
        if self._task.cancelled():
            return "cancelled"
        elif not self._task.done():
            return "running"  # asyncio doesn't distinguish between running and pending
        else:
            if exc := self.exception():
                timeout_msg = "504 Server Error: Gateway Timeout"
                if isinstance(exc, ReadTimeout) or timeout_msg in str(exc):
                    return "timeout"
                # Should be handled already but keeping for consistency with sync version
                if isinstance(exc, asyncio.CancelledError):
                    return "cancelled"
                return "error"
            return "success"


class BaseJobPool(ABC):
    """Base class for JobPool implementations"""

    def __init__(
        self,
        udf,
        arg_list,
        kwargs=None,
        engine="remote",
        max_retry=2,
        before_run=None,
        wait_sleep=0.01,
        before_submit=0.01,
    ):
        self.udf = udf
        self.arg_list = arg_list
        self.n_jobs = len(self.arg_list)
        self._kwargs = kwargs or {}
        self._engine = engine
        self._max_retry = max_retry
        self._before_run = before_run
        self._wait_sleep = wait_sleep
        self._before_submit = before_submit
        self._restarted_at = None
        self._started_at = None
        self._cancel_retry = False
        self._futures: List[BaseFuture] = []

    @abstractmethod
    def _create_future(self, index, args) -> BaseFuture:
        pass

    @abstractmethod
    def _start_jobs(self):
        pass

    @abstractmethod
    def cancel(self, wait: bool = False):
        pass

    def _get_status(self) -> List[bool]:
        return [f.done() for f in self._futures]

    def _status_counts(self) -> Dict[Status, int]:
        counts = defaultdict(int)
        for f in self._futures:
            counts[f.status()] += 1

        # TODO: always sort counts keys by status order
        return counts

    def _status_message(self) -> str:
        counts = self._status_counts()

        message_parts = []
        for status in ("running", "timeout", "error", "success", "pending"):
            c = counts[status]
            if c:
                message_parts.append(f"{c} {status}")

        if len(message_parts) == 0:
            return "empty"

        return ", ".join(message_parts)

    def _get_progress(self) -> Tuple[int, str]:
        # floor as to not show 100 if not every one is actually done
        percentage = floor(sum(self._get_status()) / self.n_jobs * 100)
        return percentage, f"{sum(self._get_status())}/{self.n_jobs}"

    def retry(self):
        """Rerun any tasks in error or timeout states. Tasks are rerun in the same pool."""
        old_futures = self._futures

        def _create_new_future(index, args):
            if old_futures[index].status() in ("error", "timeout"):
                return self._create_future(index, args)
            return old_futures[index]

        self._restarted_at = datetime.now()
        self._futures = [
            _create_new_future(index, args) for index, args in enumerate(self.arg_list)
        ]

    def total_time(self, since_retry: bool = False) -> timedelta:
        """Returns how long the entire job took.

        If only partial results are available, returns based on the last task to have been completed.
        """
        started_at = (
            self._restarted_at
            if since_retry and self._restarted_at is not None
            else self._started_at
        )
        if started_at is None:
            raise ValueError("JobPool has not been started")
        all_result_at = [f._ended_at for f in self._futures if f._ended_at is not None]
        if not len(all_result_at):
            raise ValueError("No result is available yet")
        last_result_at = max(all_result_at)
        return last_result_at - self._started_at

    def times(self) -> list[Optional[timedelta]]:
        """Time taken for each task.

        Incomplete tasks will be reported as None."""
        return [f.time() for f in self._futures]

    def done(self) -> bool:
        """True if all tasks have finished, regardless of success or failure."""
        return all(fut.done() for fut in self._futures)

    def all_succeeded(self) -> bool:
        """True if all tasks finished with success"""
        return all(f.status() == "success" for f in self._futures)

    def any_failed(self) -> bool:
        """True if any task finished with an error"""
        return any(f.status() in ("error", "timeout") for f in self._futures)

    def any_succeeded(self) -> bool:
        """True if any task finished with success"""
        return any(f.status() == "success" for f in self._futures)

    def arg_df(self):
        """The arguments passed to runs as a DataFrame"""
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        return pd.DataFrame(self.arg_list)

    def status(self):
        """Return a Series indexed by status of task counts"""
        return self.df().value_counts("status")

    def wait(self):
        """Wait until all jobs are finished

        Use fused.options.show.enable_tqdm to enable/disable tqdm.
        Use pool._wait_sleep to set if sleep should occur while waiting.
        """

        def _noop(*args, **kwargs):
            pass

        after = _noop
        update = _noop
        if OPTIONS.show.enable_tqdm:
            from tqdm.auto import tqdm

            t = tqdm(total=self.n_jobs, bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt}")

            def _tqdm_after():
                t.update(self.n_jobs - t.n)
                t.close()

            after = _tqdm_after
            update = lambda n_done: t.update(n_done - t.n)

        while not self.done():
            n_done = sum(self._get_status())
            update(n_done)
            if self._wait_sleep is not None:
                time.sleep(self._wait_sleep)

        after()

    def tail(self, stop_on_exception=False):
        """Wait until all jobs are finished, printing statuses as they become available.

        This is useful for interactively watching for the state of the pool.

        Use pool._wait_sleep to set if sleep should occur while waiting.
        """
        seen = set(self.results_now(return_exceptions=not stop_on_exception).keys())
        if len(seen):
            print(self._status_message())

            if len(seen) == len(self._futures):
                # Nothing left to tail
                return

        while len(seen) != len(self._futures):
            done = self.results_now(return_exceptions=not stop_on_exception)
            for new_key in done.keys():
                if new_key not in seen:
                    seen.add(new_key)
                    # TODO: Logger support, rather than print statement?
                    print(
                        f"[{datetime.now()}] Run #{new_key} {self._futures[new_key]._args} {self._futures[new_key].status()} ({self._status_message()})"
                    )

            if self._wait_sleep is not None:
                time.sleep(self._wait_sleep)

        print(f"End of tail\n{self._status_message()}")

    def results(self, return_exceptions=False) -> List[Any]:
        """Retrieve all results of the job.

        Results are ordered by the order of the args list."""
        results = []
        for fut in self._futures:
            try:
                results.append(fut.result())
            except Exception:
                if return_exceptions:
                    results.append(fut.exception())
                else:
                    raise
        return results

    def results_now(self, return_exceptions=False) -> Dict[int, Any]:
        """Retrieve the results that are currently done.

        Results are indexed by position in the args list."""
        results = {}
        for index, fut in enumerate(self._futures):
            if fut.done():
                try:
                    results[index] = fut.result()
                except Exception:
                    if return_exceptions:
                        results[index] = fut.exception()
                    else:
                        raise
        return results

    def df(
        self,
        status_column: Optional[str] = "status",
        result_column: Optional[str] = "result",
        time_column: Optional[str] = "time",
        logs_column: Optional[str] = "logs",
        exception_column: Optional[str] = None,
        include_exceptions: bool = True,
    ):
        """
        Get a DataFrame of results as they are currently.
        The DataFrame will have columns for each argument passed, and columns for:
        `status`, `result`, `time`, `logs` and optionally `exception`.
        """
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results_df = pd.DataFrame(self.arg_list)
        status_list = [f.status() for f in self._futures]
        if status_column:
            results_df[status_column] = status_list

        if time_column:
            time_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                t = fut.time()
                if t is not None:
                    time_list[i] = t.total_seconds()

            results_df[time_column] = pd.Series(time_list, dtype=float)

        if result_column:
            result_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    try:
                        res = fut.result()
                    except Exception:
                        res = fut.exception() if include_exceptions else None

                    result_list[i] = res

            results_df[result_column] = result_list

        if logs_column:
            logs_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    logs_list[i] = fut.logs()

            results_df[logs_column] = logs_list

        if exception_column:
            results_df[exception_column] = [fut.exception() for fut in self._futures]

        return results_df

    def get_status_df(self):
        warnings.warn(
            "the 'get_status_df()' method is deprecated, use '.df()' instead.",
            FusedDeprecationWarning,
        )
        return self.df()

    def get_results_df(self, ignore_exceptions=False):
        warnings.warn(
            "the 'get_results_df()' method is deprecated, use '.df()' instead.",
            FusedDeprecationWarning,
        )
        self.wait()
        return self.df(include_exceptions=not ignore_exceptions)

    def errors(self) -> Dict[int, Exception]:
        """Retrieve the results that are currently done and are errors.

        Results are indexed by position in the args list."""
        errors = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in STATUS_FAILED:
                errors[index] = fut.exception()
        return errors

    def first_error(self) -> Optional[Exception]:
        """Retrieve the first (by order of arguments) error result, or None."""
        for fut in self._futures:
            if fut.status() in STATUS_FAILED:
                return fut.exception()

        return None

    def logs(self) -> list[Optional[str]]:
        """Logs for each task.

        Incomplete tasks will be reported as None."""
        return [
            (f.logs() if f.status() not in STATUS_NOT_FINISHED else None)
            for f in self._futures
        ]

    def first_log(self) -> Optional[str]:
        """Retrieve the first (by order of arguments) logs, or None."""
        for f in self._futures:
            if f.status() not in STATUS_NOT_FINISHED:
                return f.logs()

    def success(self) -> Dict[int, Any]:
        """Retrieve the results that are currently done and are successful.

        Results are indexed by position in the args list."""
        success = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("success",):
                success[index] = fut.result()
        return success

    def pending(self) -> Dict[int, Any]:
        """Retrieve the arguments that are currently pending and not yet submitted."""
        pending = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("pending",):
                pending[index] = fut._args
        return pending

    def running(self) -> Dict[int, Any]:
        """Retrieve the results that are currently running."""
        running = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("running",):
                running[index] = fut._args
        return running

    def cancelled(self) -> Dict[int, Any]:
        """Retrieve the arguments that were cancelled and not run."""
        cancelled = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("cancelled",):
                cancelled[index] = fut._args
        return cancelled

    def collect(self, ignore_exceptions=False, flatten=True):
        """Collect all results into a DataFrame"""
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        # Handles both thread and async
        self.wait()
        results = self.results(return_exceptions=ignore_exceptions)
        mask = [not isinstance(r, Exception) for r in results]
        results = dict(
            res
            for res in zip(range(self.n_jobs), results)
            if not isinstance(res[1], Exception)
        )

        results_pandas = all(isinstance(res, pd.DataFrame) for res in results.values())
        if results_pandas and len(results) and flatten:
            # TODO: flatten can un-occur depending on results, this seems wrong
            results_df = pd.concat(results)
        else:
            results_df = pd.DataFrame({"result": pd.Series(results)})
            results_df.index = pd.MultiIndex.from_product([results_df.index])

        args_df = pd.DataFrame(self.arg_list)
        args_df = args_df[mask]

        _coerce_hashable(args_df)

        # combine concatenated results with arguments as prepended index levels
        assert len(results_df.index.levels[0]) == len(args_df)
        args_index = pd.MultiIndex.from_frame(args_df)
        args_codes = [c.take(results_df.index.codes[0]) for c in args_index.codes]
        new_idx = pd.MultiIndex(
            levels=list(args_index.levels) + results_df.index.levels[1:],
            codes=args_codes + results_df.index.codes[1:],
            names=list(args_df.columns) + results_df.index.names[1:],
        )
        if new_idx.nlevels == 1:
            new_idx = new_idx.get_level_values(0)
        results_df.index = new_idx
        return results_df

    def __getitem__(self, idx: int) -> BaseFuture:
        return self._futures[idx]

    def __len__(self) -> int:
        return self.n_jobs

    def __repr__(self) -> str:
        return f"<JobPool with {self.n_jobs} jobs [{self._status_message()}]>"

    def _repr_html_(self) -> str:
        # TODO we could provide a more informative repr in notebooks (e.g. showing
        # a table of the individual jobs and their status?)
        counts = self._status_counts()
        status_table_data = [
            f"<tr><td>{status}</td><td>{count}</td></tr>"
            for status, count in counts.items()
        ]
        return f"<table><tr><th>Status</th><th>Count</th></tr>{''.join(status_table_data)}</table>"


class JobPool(BaseJobPool):
    """Thread-based JobPool implementation using ThreadPoolExecutor"""

    def __init__(
        self,
        udf,
        arg_list,
        kwargs=None,
        engine="remote",
        max_workers=None,
        max_retry=2,
        before_run=None,
        wait_sleep=0.01,
        before_submit=0.01,
    ):
        super().__init__(
            udf,
            arg_list,
            kwargs,
            engine,
            max_retry,
            before_run,
            wait_sleep,
            before_submit,
        )
        self._pool = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)

    def _create_future(self, index, args) -> Future:
        future = Future(None, index, args)

        def _run(args):
            # currently we have to add a small delay between starting the UDF runs
            # to avoid overloading the server
            if self._before_run is not None:
                time.sleep(self._before_run)

            try:
                future._started_at = datetime.now()

                return fused_run(
                    self.udf,
                    engine=self._engine,
                    _return_response=True,
                    max_retry=self._max_retry,
                    _cancel_callback=lambda: self._cancel_retry,
                    **args,
                    **self._kwargs,
                )
            except Exception as exc:
                # ReadTime or HTTPError can happen on time-out or other server error
                return exc
            finally:
                future._ended_at = datetime.now()

        if self._before_submit:
            time.sleep(self._before_submit)

        future._future = self._pool.submit(_run, args)

        return future

    def _start_jobs(self):
        self._started_at = datetime.now()
        self._futures = [
            self._create_future(index, args) for index, args in enumerate(self.arg_list)
        ]

    def cancel(self, wait: bool = False):
        """Cancel any pending (not running) tasks.

        Note it will not be possible to retry on the same JobPool later."""
        # Signal running tasks to stop retrying
        self._cancel_retry = True

        self._pool.shutdown(wait=wait, cancel_futures=True)
        counts = self._status_counts()

        if not wait:
            message = f"{counts['cancelled']} task(s) cancelled successfully"
            if counts["running"]:
                message += f", {counts['running']} task(s) already in progress and can not be cancelled."
            print(message)


class AsyncJobPool(BaseJobPool):
    """Async-based JobPool implementation using asyncio"""

    def __init__(
        self,
        udf,
        arg_list,
        kwargs=None,
        engine="remote",
        max_workers=None,
        max_retry=2,
        before_run=None,
        wait_sleep=0.01,
        before_submit=0.01,
    ):
        super().__init__(
            udf,
            arg_list,
            kwargs,
            engine,
            max_retry,
            before_run,
            wait_sleep,
            before_submit,
        )
        self._semaphore = asyncio.Semaphore(max_workers)

    def __repr__(self) -> str:
        return f"<AsyncJobPool with {self.n_jobs} jobs [{self._status_message()}]>"

    def check_loop(self):
        try:
            _ = asyncio.get_running_loop()
            if not HAS_NEST_ASYNCIO:
                raise ImportError(
                    "nest_asyncio is required to use AsyncJobPool in environments with an existing event loop "
                    "(like Jupyter notebooks). Install it with: pip install nest_asyncio"
                )
        except RuntimeError:
            # No event loop running, which is fine
            pass

    async def _create_future(self, index, args) -> AsyncFuture:
        future = AsyncFuture(None, index, args)

        async def _run(args):
            async with self._semaphore:
                # currently we have to add a small delay between starting the UDF runs
                # to avoid overloading the server
                if self._before_run is not None:
                    await asyncio.sleep(self._before_run)

                try:
                    future._started_at = datetime.now()

                    result = await fused_run_async(
                        self.udf,
                        engine=self._engine,
                        _return_response=True,
                        max_retry=self._max_retry,
                        _cancel_callback=lambda: self._cancel_retry,
                        **args,
                        **self._kwargs,
                    )
                    return result
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    return e
                finally:
                    future._ended_at = datetime.now()

        if self._before_submit:
            await asyncio.sleep(self._before_submit)

        future._task = asyncio.create_task(_run(args))

        return future

    async def _start_jobs(self):
        self._started_at = datetime.now()
        self._futures = []

        for index, args in enumerate(self.arg_list):
            future = await self._create_future(index, args)
            self._futures.append(future)

    def cancel(self, wait: bool = False):
        """Cancel any pending (not running) tasks."""
        self.check_loop()

        return asyncio.run(self.cancel_async(wait))

    async def cancel_async(self, wait: bool = False):
        self._cancel_retry = True

        for future in self._futures:
            if not future._task.done():
                future._task.cancel()

        counts = self._status_counts()

        if not wait:
            message = f"{counts['cancelled']} task(s) cancelled successfully"
            if counts["running"]:
                message += f", {counts['running']} task(s) already in progress and can not be cancelled."
            print(message)
        else:
            await asyncio.gather(
                *(f._task for f in self._futures), return_exceptions=True
            )

    async def _wait_with_progress_async(self, update_progress):
        """Async wait with incremental progress updates"""
        if not self._futures:
            return

        while not self.done():
            n_done = sum(1 for f in self._futures if f.done())
            update_progress(n_done)

            if self._wait_sleep is not None:
                await asyncio.sleep(self._wait_sleep)
            else:
                # Small sleep to prevent busy waiting
                await asyncio.sleep(0.01)

    async def results_now_async(self, return_exceptions=False) -> Dict[int, Any]:
        """Async version of results_now that doesn't block the event loop"""
        results = {}
        for index, fut in enumerate(self._futures):
            if fut.done():
                try:
                    results[index] = fut.result()
                except Exception:
                    if return_exceptions:
                        results[index] = fut.exception()
                    else:
                        raise
        return results

    async def results_async(self, return_exceptions=False) -> List[Any]:
        """Async version of results that assumes waiting has already been done"""
        results = []
        for fut in self._futures:
            try:
                results.append(fut.result())
            except Exception:
                if return_exceptions:
                    results.append(fut.exception())
                else:
                    raise
        return results

    async def tail_async(self, stop_on_exception=False):
        """Async version of tail that doesn't block the event loop"""
        seen = set(
            (
                await self.results_now_async(return_exceptions=not stop_on_exception)
            ).keys()
        )
        if len(seen):
            print(self._status_message())

            if len(seen) == len(self._futures):
                # Nothing left to tail
                return

        while len(seen) != len(self._futures):
            done = await self.results_now_async(return_exceptions=not stop_on_exception)
            for new_key in done.keys():
                if new_key not in seen:
                    seen.add(new_key)
                    print(
                        f"[{datetime.now()}] Run #{new_key} {self._futures[new_key]._args} {self._futures[new_key].status()} ({self._status_message()})"
                    )

            if self._wait_sleep is not None:
                await asyncio.sleep(self._wait_sleep)

        print(f"End of tail\n{self._status_message()}")

    def tail(self, stop_on_exception=False):
        """Override tail to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.tail_async(stop_on_exception))

    def results_now(self, return_exceptions=False) -> Dict[int, Any]:
        """Override results_now to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.results_now_async(return_exceptions))

    def results(self, return_exceptions=False) -> List[Any]:
        """Override results to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.results_async(return_exceptions))

    async def df_async(
        self,
        status_column: Optional[str] = "status",
        result_column: Optional[str] = "result",
        time_column: Optional[str] = "time",
        logs_column: Optional[str] = "logs",
        exception_column: Optional[str] = None,
        include_exceptions: bool = True,
    ):
        """Async version of df() that doesn't block the event loop"""
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results_df = pd.DataFrame(self.arg_list)
        status_list = [f.status() for f in self._futures]
        if status_column:
            results_df[status_column] = status_list

        if time_column:
            time_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                t = fut.time()
                if t is not None:
                    time_list[i] = t.total_seconds()

            results_df[time_column] = pd.Series(time_list, dtype=float)

        if result_column:
            result_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    try:
                        res = fut.result()
                    except Exception:
                        res = fut.exception() if include_exceptions else None

                    result_list[i] = res

            results_df[result_column] = result_list

        if logs_column:
            logs_list = [None] * len(self._futures)
            for i, fut in enumerate(self._futures):
                if status_list[i] not in STATUS_NOT_FINISHED:
                    logs_list[i] = fut.logs()

            results_df[logs_column] = logs_list

        if exception_column:
            results_df[exception_column] = [fut.exception() for fut in self._futures]

        return results_df

    def df(
        self,
        status_column: Optional[str] = "status",
        result_column: Optional[str] = "result",
        time_column: Optional[str] = "time",
        logs_column: Optional[str] = "logs",
        exception_column: Optional[str] = None,
        include_exceptions: bool = True,
    ):
        """Override df to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(
            self.df_async(
                status_column,
                result_column,
                time_column,
                logs_column,
                exception_column,
                include_exceptions,
            )
        )

    async def success_async(self) -> Dict[int, Any]:
        """Async version of success that doesn't block the event loop"""
        success = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in ("success",):
                success[index] = fut.result()
        return success

    def success(self) -> Dict[int, Any]:
        """Override success to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.success_async())

    async def errors_async(self) -> Dict[int, Exception]:
        """Async version of errors that doesn't block the event loop"""
        errors = {}
        for index, fut in enumerate(self._futures):
            if fut.status() in STATUS_FAILED:
                errors[index] = fut.exception()
        return errors

    def errors(self) -> Dict[int, Exception]:
        """Override errors to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.errors_async())

    async def first_error_async(self) -> Optional[Exception]:
        """Async version of first_error that doesn't block the event loop"""
        for fut in self._futures:
            if fut.status() in STATUS_FAILED:
                return fut.exception()
        return None

    def first_error(self) -> Optional[Exception]:
        """Override first_error to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.first_error_async())

    async def logs_async(self) -> list[Optional[str]]:
        """Async version of logs that doesn't block the event loop"""
        return [
            (f.logs() if f.status() not in STATUS_NOT_FINISHED else None)
            for f in self._futures
        ]

    def logs(self) -> list[Optional[str]]:
        """Override logs to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.logs_async())

    async def first_log_async(self) -> Optional[str]:
        """Async version of first_log that doesn't block the event loop"""
        for f in self._futures:
            if f.status() not in STATUS_NOT_FINISHED:
                return f.logs()
        return None

    def first_log(self) -> Optional[str]:
        """Override first_log to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.first_log_async())

    def status(self):
        """Override status to handle async context properly"""
        self.check_loop()

        # Either a new loop is created or we are able to nest
        return asyncio.run(self.df_async()).value_counts("status")

    def wait(self):
        """Wait until all jobs are finished (sync wrapper around wait_async)"""
        if not self._futures:
            return

        self.check_loop()

        # Set up progress tracking (same as BaseJobPool.wait)
        def _noop(*args, **kwargs):
            pass

        after = _noop
        update = _noop
        if OPTIONS.show.enable_tqdm:
            from tqdm.auto import tqdm

            t = tqdm(total=self.n_jobs, bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt}")

            def _tqdm_after():
                t.update(self.n_jobs - t.n)
                t.close()

            after = _tqdm_after
            update = lambda n_done: t.update(n_done - t.n)

        # Run async wait with progress updates
        asyncio.run(self._wait_with_progress_async(update))
        after()

    async def wait_async(self):
        if not self._futures:
            return

        self.check_loop()

        # Set up progress tracking (same as BaseJobPool.wait)
        def _noop(*args, **kwargs):
            pass

        after = _noop
        update = _noop
        if OPTIONS.show.enable_tqdm:
            from tqdm.auto import tqdm

            t = tqdm(total=self.n_jobs, bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt}")

            def _tqdm_after():
                t.update(self.n_jobs - t.n)
                t.close()

            after = _tqdm_after
            update = lambda n_done: t.update(n_done - t.n)

        await self._wait_with_progress_async(update)
        after()

    async def _retry_async(self):
        old_futures = self._futures

        async def _create_new_future(index, args):
            if old_futures[index].status() in ("error", "timeout"):
                return await self._create_future(index, args)
            return old_futures[index]

        self._restarted_at = datetime.now()
        self._futures = []
        for index, args in enumerate(self.arg_list):
            future = await _create_new_future(index, args)
            self._futures.append(future)

    def retry(self):
        self.check_loop()

        # Either a new loop is created or we are able to nest
        asyncio.run(self._retry_async())


def _validate_arg_list(arg_list, udf):
    if not len(arg_list):
        raise ValueError("arg_list must be a non-empty")

    if HAS_PANDAS and isinstance(arg_list, PD_DATAFRAME):
        return arg_list.to_dict(orient="records")

    if not isinstance(arg_list[0], dict):
        if not isinstance(udf, Udf):
            raise ValueError(
                "arg_list must be a list of dictionaries. A simple list to pass "
                "as first positional argument is only supported for UDF objects."
            )
        if udf._parameter_list is None:
            # TODO: Run _parameter_list detection here
            raise NotImplementedError(
                "arg_list must be a list of dictionaries. Could not detect the first "
                "positional argument name."
            )
        if len(udf._parameter_list) == 0:
            raise ValueError(
                "arg_list must be a list of dictionaries. UDF does not accept any arguments."
            )
        name = udf._parameter_list[0]
        arg_list = [{name: arg} for arg in arg_list]

    return arg_list


@overload
def submit(
    udf: AnyBaseUdf | FunctionType | str,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[False] = False,
    collect: Literal[True] = True,
    execution_type: ExecutionType = "thread_pool",
    **kwargs,
) -> "pd.DataFrame": ...


@overload
def submit(
    udf: AnyBaseUdf | FunctionType | str,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[False] = False,
    collect: Literal[False] = False,
    execution_type: ExecutionType = "thread_pool",
    **kwargs,
) -> Union[JobPool, AsyncJobPool]: ...


@overload
def submit(
    udf: AnyBaseUdf | FunctionType | str,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: Literal[True] = True,
    collect: Literal[True] = True,
    execution_type: ExecutionType = "thread_pool",
    **kwargs,
) -> ResultType: ...


def submit(
    udf: AnyBaseUdf | FunctionType | str,
    arg_list,
    /,
    *,
    engine: Optional[Literal["remote", "local"]] = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: bool = False,
    collect: bool = True,
    execution_type: ExecutionType = "thread_pool",
    cache_max_age: Optional[str] = None,
    cache: bool = True,
    ignore_exceptions: bool = False,
    flatten: bool = True,
    _before_run: Optional[float] = None,
    _before_submit: Optional[float] = 0.01,
    **kwargs,
) -> Union[BaseJobPool, ResultType, "pd.DataFrame"]:
    """
    Executes a user-defined function (UDF) multiple times for a list of input
    parameters, and return immediately a "lazy" JobPool object allowing
    to inspect the jobs and wait on the results.

    Each individual UDF run will be cached following the standard caching logic as with `fused.run()`
    and the specified `cache_max_age`. Additionally, when `collect=True` (the default), the collected results
    are cached locally for the duration of `cache_max_age` or 12h by default.

    See `fused.run` for more details on the UDF execution.

    Args:
        udf: the UDF to execute.
            See `fused.run` for more details on how to specify the UDF.
        arg_list: a list of input parameters for the UDF. Can be specified as:
            - a list of values for parametrizing over a single parameter, i.e.
              the first parameter of the UDF
            - a list of dictionaries for parametrizing over multiple parameters
            - A DataFrame for parametrizing over multiple parameters where each
              row is a set of parameters

        engine: The execution engine to use. Defaults to 'remote'.
        max_workers: The maximum number of workers to use. Defaults to 32.
        max_retry: The maximum number of retries for failed jobs. Defaults to 2.
        debug_mode: If True, executes only the first item in arg_list directly using
            `fused.run()`, useful for debugging UDF execution. Default is False.
        collect: If True, waits for all jobs to complete and returns the collected DataFrame
            containing the results. If False, returns a JobPool object, which is non-blocking
            and allows you to inspect the individual results and logs.
            Default is True.
        execution_type: The type of batching to use. Either "thread_pool" (default) for
            ThreadPoolExecutor-based concurrency or "async_loop" for asyncio-based concurrency.
        cache_max_age: The maximum age when returning a result from the cache.
            Supported units are seconds (s), minutes (m), hours (h), and days (d)
            (e.g. "48h", "10s", etc.).
            Default is `None` so a UDF run with `fused.run()` will follow
            `cache_max_age` defined in `@fused.udf()` unless this value is changed.
        cache: Set to False as a shortcut for `cache_max_age='0s'` to disable caching.
        ignore_exceptions: Set to True to ignore exceptions when collecting results.
            Runs that result in exceptions will be silently ignored. Defaults to False.
        flatten: Set to True to receive a DataFrame of results, without nesting of a
            `results` column, when collecting results. When False, results will be nested
            in a `results` column. If the UDF does not return a DataFrame (e.g. a string
            instead,) results will be nested in a `results` column regardless of this setting.
            Defaults to True.
        **kwargs: Additional (constant) keyword arguments to pass to the UDF.

    Returns:
        JobPool, AsyncJobPool, or DataFrame depending on execution_type and collect parameters

    Examples:
        Run a UDF multiple times for the values 0 to 9 passed to as the first
        positional argument of the UDF:
        ```py
        df = fused.submit("username@fused.io/my_udf_name", range(10))
        ```

        Using async batch type:
        ```py
        df = fused.submit(udf, range(10), execution_type="async_loop")
        ```

        Being explicit about the parameter name:
        ```py
        df = fused.submit(udf, [dict(n=i) for i in range(10)])
        ```

        Get the pool of ongoing tasks:
        ```py
        pool = fused.submit(udf, [dict(n=i) for i in range(10)], collect=False)
        ```

    """
    if execution_type not in ("thread_pool", "async_loop"):
        raise ValueError(
            f"'execution_type' must be 'thread_pool' or 'async_loop', got {execution_type}"
        )

    if isinstance(udf, FunctionType):
        if udf.__name__ == "<lambda>":
            # This will not work correctly in fused.udf. If we find a way to parse just the AST
            # of the lambda (without any surrounding function call, assignment, etc.) then we can
            # support lambda here
            raise TypeError(
                """Lambda expressions cannot be passed into fused.submit(). Create a function with @fused.udf instead:
@fused.udf
def udf(x):
    return x

fused.submit(udf, arg_list)
"""
            )
        # TODO: Move this logic to fused.run?
        udf: Udf = fused_udf(udf)

    arg_list = _validate_arg_list(arg_list, udf)

    if cache_max_age is not None:
        kwargs["cache_max_age"] = cache_max_age
    if not cache:
        kwargs["cache"] = cache

    if debug_mode:
        if not collect:
            warnings.warn(
                "'debug_mode=True' and 'collect=False' are mutually exclusive (the "
                "result of the first run is always returned directly), and the collect "
                "keyword is ignored in this case."
            )
        return fused_run(udf, engine=engine, **arg_list[0], **kwargs)

    max_workers = min(max_workers or 32, 1024)

    # set the cache_max_age if not provided, to be used for caching the submit collect results
    submit_cache_max_age = _parse_ttl(cache_max_age, cache=cache)
    if submit_cache_max_age is None:
        if isinstance(udf, BaseUdf) and udf.cache_max_age is not None:
            submit_cache_max_age = udf.cache_max_age
        elif isinstance(udf, str):
            submit_cache_max_age = 0
        else:
            submit_cache_max_age = _parse_ttl(DEFAULT_CACHE_MAX_AGE)

    if collect and submit_cache_max_age != 0:
        udf_name = udf if isinstance(udf, str) else udf.name or udf.entrypoint

        # cache the submit collect results
        fn = fused_cache(
            _submit_internal,
            cache_max_age=submit_cache_max_age,
            cache_folder_path=f"submit-{udf_name}",
            cache_reset=kwargs.get("cache_reset", None),
            cache_verbose=kwargs.get("cache_verbose", None),
            cache_storage=kwargs.get("cache_storage", "auto"),
        )
    else:
        fn = _submit_internal

    return fn(
        udf,
        arg_list,
        engine=engine,
        max_workers=max_workers,
        max_retry=max_retry,
        collect=collect,
        execution_type=execution_type,
        ignore_exceptions=ignore_exceptions,
        flatten=flatten,
        _before_run=_before_run,
        _before_submit=_before_submit,
        **kwargs,
    )


def _submit_internal(
    udf: AnyBaseUdf | str,
    arg_list: Any,
    /,
    *,
    engine: Literal["remote", "local"],
    max_workers: int,
    max_retry: int,
    collect: bool,
    execution_type: ExecutionType,
    ignore_exceptions: bool,
    flatten: bool,
    _before_run: float | None,
    _before_submit: float,
    **kwargs,
):
    """
    Helper function that can be cached conditionally
    """
    if execution_type == "async_loop":
        if HAS_NEST_ASYNCIO:
            nest_asyncio.apply()
        else:
            raise ImportError("async_loop execution_type requires 'nest_asyncio'")

        job_pool = AsyncJobPool(
            udf,
            arg_list,
            kwargs,
            engine=engine,
            max_workers=max_workers,
            max_retry=max_retry,
            before_run=_before_run,
            before_submit=_before_submit,
        )

        async def _run_async_batch():
            await job_pool._start_jobs()
            if collect:
                return job_pool.collect(
                    ignore_exceptions=ignore_exceptions, flatten=flatten
                )
            return job_pool

        # Either loop exists or needs to be created.
        return asyncio.run(_run_async_batch())

    else:
        job_pool = JobPool(
            udf,
            arg_list,
            kwargs,
            engine=engine,
            max_workers=max_workers,
            max_retry=max_retry,
            before_run=_before_run,
            before_submit=_before_submit,
        )
        job_pool._start_jobs()
        if collect:
            return job_pool.collect(
                ignore_exceptions=ignore_exceptions, flatten=flatten
            )
        return job_pool



from __future__ import annotations

import asyncio
import json
import sys
import time
from typing import TYPE_CHECKING, Any

import aiohttp
import requests
import yarl

from fused._optional_deps import (
    GPD_GEODATAFRAME,
    HAS_GEOPANDAS,
    HAS_PANDAS,
    HAS_SHAPELY,
    PD_DATAFRAME,
    SHAPELY_GEOMETRY,
)
from fused._options import options as OPTIONS
from fused._request import raise_for_status, raise_for_status_async
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.types import UdfRuntimeError

from ..models import AnyJobStepConfig
from ..models.udf import AnyBaseUdf
from ..models.udf._eval_result import is_response_cached
from ._impl._context_impl import (
    context_get_auth_header,
    context_get_user_email,
    context_get_user_email_async,
)
from ._impl._realtime_ops_impl import (
    get_recursion_factor,
    make_realtime_url,
    make_shared_realtime_url,
)
from ._serialization import (
    deserialize_html,
    deserialize_json,
    deserialize_npy,
    deserialize_parquet,
    deserialize_png,
    deserialize_tiff,
    deserialize_zip,
)

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr


def serialize_realtime_params(params: dict[str, Any]):
    result = {}
    for param_name, param_value in params.items():
        if (HAS_GEOPANDAS and isinstance(param_value, GPD_GEODATAFRAME)) or (
            HAS_PANDAS and isinstance(param_value, PD_DATAFRAME)
        ):
            result[param_name] = param_value.to_json()
        elif isinstance(param_value, (list, tuple, dict, bool)):
            result[param_name] = json.dumps(param_value)
        elif HAS_SHAPELY and isinstance(param_value, SHAPELY_GEOMETRY):
            import shapely

            result[param_name] = shapely.to_wkt(param_value)
        else:
            result[param_name] = param_value

    return result


async def _realtime_raise_for_status_async(r: aiohttp.ClientResponse):
    if r.status >= 400 and "x-fused-error" in r.headers:
        msg = str(r.headers["x-fused-error"])
        raise aiohttp.ClientResponseError(
            request_info=r.request_info,
            status=r.status,
            message=msg,
            headers=r.headers,
            history=r.history,
        )
    if r.status >= 400 and "x-fused-metadata" in r.headers:
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        # stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        # has_exception = fused_metadata.get("has_exception", False)

        # Extract error line, if exists.
        # error_lineno = fused_metadata.get("lineno")

        error_message = ""
        if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
            error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
        elif "exception" in fused_metadata and fused_metadata["exception"]:
            error_message = fused_metadata["exception"]

        if stdout:
            sys.stdout.write(stdout)
        raise UdfRuntimeError(error_message, child_exception_class=exception_class)

    await raise_for_status_async(r)


async def _realtime_follow_redirect_async(
    *, session: aiohttp.ClientSession, r: aiohttp.ClientResponse
):
    if r.status >= 300 and r.status < 400 and "location" in r.headers:
        # Per this link, aiohttp will mangle the redirect URL
        # https://stackoverflow.com/questions/77319421/aiohttp-showing-403-forbidden-error-but-requests-get-giving-200-ok-response
        url = yarl.URL(r.headers["location"], encoded=True)
        return await session.get(url)
    return r


def run_tile(
    email_or_team: str,
    id: str | None = None,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _client_id: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Executes a private tile-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to run a UDF on a specific tile defined by its x, y, and z coordinates, and
    sends a request to the server. It supports customization of the output data types for vector and raster data,
    as well as additional parameters for the UDF execution.

    Args:
        email (str): Email address of user account associated with the UDF.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a pre-defined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a pre-defined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF on the specified tile.
    """
    if id is None:
        id = email_or_team
        email_or_team = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email_or_team}/{id}/tiles/{z}/{x}/{y}"
    return _run_and_process(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


def run_shared_tile(
    token: str,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _client_id: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Executes a shared tile-based UDF.

    This function constructs a URL to run a UDF on a specific tile defined by its x, y, and z coordinates, and
    sends a request to the server. It supports customization of the output data types for vector and raster data,
    as well as additional parameters for the UDF execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a pre-defined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a pre-defined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF on the specified tile.
    """
    url = f"{make_shared_realtime_url(token)}/run/tiles/{z}/{x}/{y}"
    return _run_and_process(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _client_id=_client_id,
        _is_shared=True,
        **params,
    )


def _run_and_process(
    url: str,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _profile: bool = False,
    _is_shared: bool = False,
    **params,
):
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector or OPTIONS.default_dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster or OPTIONS.default_dtype_out_raster,
        "profile": _profile,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    time_start = time.perf_counter()
    r = requests.get(
        url=url,
        params=req_params,
        headers={
            **access_token_header,
            **(OPTIONS.default_run_headers or {}),
            "Fused-Recursion": f"{recursion_factor}",
        },
    )
    time_end = time.perf_counter()

    time_taken_seconds = time_end - time_start
    return _process_response(
        r, step_config=None, time_taken_seconds=time_taken_seconds, profile=_profile
    )


def run_file(
    email_or_team: str,
    id: str | None = None,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _client_id: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Executes a private file-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to run a UDF associated with the given email and ID, allowing for output data type customization for both vector and raster outputs. It also supports additional parameters for the UDF execution.

    Args:
        email (str): Email address of user account associated with the UDF.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        The response from the server after executing the UDF.
    """
    if id is None:
        id = email_or_team
        email_or_team = context_get_user_email()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email_or_team}/{id}"
    return _run_and_process(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


def run_shared_file(
    token: str,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Executes a shared file-based UDF.

    This function constructs a URL for running an operation on a file accessible via a shared token. It allows for customization of the output data types for vector and raster data and supports additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        The response from the server after executing the operation on the file.

    Raises:
        Exception: Describes various exceptions that could occur during the function execution, including but not limited to invalid parameters, network errors, unauthorized access errors, or server-side errors.

    Note:
        This function is designed to access shared operations that require a token for authorization. It requires network access to communicate with the server hosting these operations and may incur data transmission costs or delays depending on the network's performance.
    """
    url = f"{make_shared_realtime_url(token)}/run/file"
    return _run_and_process(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )


async def run_tile_async(
    email_or_team: str,
    id: str | None = None,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _client_id: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Asynchronously executes a private tile-based UDF indexed under the specified email and ID. The calling user must have the necessary permissions to execute the UDF.

    This function constructs a URL to asynchronously run a UDF on a specific tile defined by its x, y, and z coordinates. It supports customization of the output data types for vector and raster data, and accommodates additional parameters for the UDF execution.

    Args:
        email (str): The user's email address, used to identify the user's saved UDFs. If the ID is not provided, this email will also be used as the ID.
        id (Optional[str]): Unique identifier for the UDF. If None, the user's email is used as the ID.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output. Defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output. Defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the UDF on the specified tile and returns the server's response. The format and content of the response depend on the UDF's implementation and the server's response format.
    """
    if id is None:
        id = email_or_team
        email_or_team = await context_get_user_email_async()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email_or_team}/{id}/tiles/{z}/{x}/{y}"
    return await _run_and_process_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


async def run_shared_tile_async(
    token: str,
    *,
    x: int,
    y: int,
    z: int,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Asynchronously executes a shared tile-based UDF using a specific access token.

    This function constructs a URL for running an operation on a tile, defined by its x, y, and z coordinates, accessible via a shared token. It allows for customization of the output data types for vector and raster data and supports additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation on the specified tile.
        x (int): The x coordinate of the tile.
        y (int): The y coordinate of the tile.
        z (int): The zoom level of the tile.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the operation on the specified tile and returns the server's response. The format and content of the response depend on the operation's implementation and the server's response format.
    """
    url = f"{make_shared_realtime_url(token)}/run/tiles/{z}/{x}/{y}"
    return await _run_and_process_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )


async def _run_and_process_async(
    url: str,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _is_shared: bool = False,
    **params,
) -> UdfEvaluationResult:
    access_token_header = context_get_auth_header(missing_ok=_is_shared)
    recursion_factor = get_recursion_factor()

    req_params = {
        # TODO...
        **({"cache_max_age": int(cache_max_age)} if cache_max_age is not None else {}),
        "dtype_out_vector": _dtype_out_vector or OPTIONS.default_dtype_out_vector,
        "dtype_out_raster": _dtype_out_raster or OPTIONS.default_dtype_out_raster,
        **(serialize_realtime_params(params) if params is not None else {}),
    }

    time_start = time.perf_counter()
    if OPTIONS.pyodide_async_requests:
        import pyodide.http

        url_with_params = yarl.URL(url, encoded=True).with_query(req_params)
        r = await pyodide.http.pyfetch(
            str(url_with_params),
            headers={
                **access_token_header,
                **(OPTIONS.default_run_headers or {}),
                "Fused-Recursion": f"{recursion_factor}",
            },
        )
        time_end = time.perf_counter()
        redirect = r.headers.get("x-fused-redirect", None)
        if redirect:
            url = yarl.URL(redirect, encoded=True)
            r = await pyodide.http.pyfetch(url)

        await _realtime_raise_for_status_async(r)

        time_taken_seconds = time_end - time_start
        result = await _process_response_async(
            r, step_config=None, time_taken_seconds=time_taken_seconds
        )

    else:
        # Use shared session for connection pooling and reuse
        session = await _get_shared_session()

        async with session.get(
            url=url,
            params=req_params,
            headers={
                **access_token_header,
                **(OPTIONS.default_run_headers or {}),
                "Fused-Recursion": f"{recursion_factor}",
            },
            allow_redirects=False,
        ) as r:
            time_end = time.perf_counter()
            r = await _realtime_follow_redirect_async(session=session, r=r)
            await _realtime_raise_for_status_async(r)

            time_taken_seconds = time_end - time_start
            result = await _process_response_async(
                r, step_config=None, time_taken_seconds=time_taken_seconds
            )

    return result


async def run_file_async(
    email_or_team: str,
    id: str | None = None,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _client_id: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Asynchronously executes a file-based UDF associated with the specific email and ID.

    This function constructs a URL to run a UDF on a server, allowing for output data type customization for vector and raster outputs and supporting additional parameters for the UDF execution. If no ID is provided, the user's email is used as the identifier.

    Args:
        email_or_team (str): The user's email address, used to identify the user's saved UDFs. If the ID is not provided, this email will also be used as the ID.
        id (Optional[str]): Unique identifier for the UDF. If None, the function fetches the user's email as the ID.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        _client_id (Optional[str]): Client identifier for API usage. If None, a default or global client ID may be used.
        **params: Additional keyword arguments for the UDF execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the UDF and returns the server's response. The format and content of the response depend on the UDF's implementation and the server's response format.
    """
    if id is None:
        id = email_or_team
        email_or_team = await context_get_user_email_async()

    url = f"{make_realtime_url(_client_id)}/api/v1/run/udf/saved/{email_or_team}/{id}"

    return await _run_and_process_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        **params,
    )


async def run_shared_file_async(
    token: str,
    *,
    cache_max_age: int | None = None,
    _dtype_out_vector: str | None = None,
    _dtype_out_raster: str | None = None,
    _run_id: str | None = None,
    **params,
) -> pd.DataFrame | xr.Dataset | None:
    """
    Asynchronously executes a shared file-based UDF using the specific access token.

    Constructs a URL to run an operation on a file accessible via a shared token, enabling customization of the output data types for vector and raster data. It accommodates additional parameters for the operation's execution.

    Args:
        token (str): A shared access token that authorizes the operation.
        cache_max_age (Optional[int]): The maximum age when returning a result from the cache.
        _dtype_out_vector (str): Desired data type for vector output, defaults to a predefined type.
        _dtype_out_raster (str): Desired data type for raster output, defaults to a predefined type.
        **params: Additional keyword arguments for the operation execution.

    Returns:
        A coroutine that, when awaited, sends a request to the server to execute the operation on the file and returns the server's response. The format and content of the response depend on the operation's implementation and the server's response format.
    """
    url = f"{make_shared_realtime_url(token)}/run/file"
    return await _run_and_process_async(
        url=url,
        cache_max_age=cache_max_age,
        _dtype_out_vector=_dtype_out_vector,
        _dtype_out_raster=_dtype_out_raster,
        _is_shared=True,
        **params,
    )


def _extract_fused_metadata(headers):
    # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
    _fused_metadata = headers.get("x-fused-metadata")
    fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

    meta = {}
    meta["profile"] = fused_metadata.get("profile")

    # Extract stdout/stderr.
    meta["stdout"] = fused_metadata.get("stdout")
    meta["stderr"] = fused_metadata.get("stderr")

    # Extract exception details
    meta["exception_class"] = fused_metadata.get("exception_class")
    meta["has_exception"] = fused_metadata.get("has_exception", False)

    # Extract error line, if exists.
    meta["error_lineno"] = fused_metadata.get("lineno")

    meta["errormsg"] = fused_metadata.get("errormsg", "")
    meta["exception"] = fused_metadata.get("exception", "")
    meta["error_type"] = fused_metadata.get("error_type")

    return meta


def _process_response(
    r: requests.Response,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
    profile: bool = False,
) -> UdfEvaluationResult:
    result_content: bytes | None = None
    udf: AnyBaseUdf | None = None
    if step_config is not None:
        udf = step_config.udf

    meta = _extract_fused_metadata(r.headers)

    if profile:
        return meta["profile"]

    if r.status_code == 200:
        result_content = r.content

        # If the UDF returned None.
        if len(result_content) == 0:
            data = None

        # Else, process response output.
        elif r.headers["content-type"] == "image/png":
            data = deserialize_png(result_content)

        elif r.headers["content-type"] == "image/tiff":
            # TODO: Automatically display tiff?
            data = deserialize_tiff(r.content)

        elif r.headers["content-type"] == "application/json":
            data = deserialize_json(result_content)

        elif r.headers["content-type"].startswith("text/html"):
            data = deserialize_html(result_content)

        elif r.headers["content-type"] == "application/zip":
            data = deserialize_zip(result_content)

        elif r.headers["content-type"] == "application/x-numpy-data":
            data = deserialize_npy(result_content)
        else:
            # assume parquet
            data = deserialize_parquet(result_content)

        cache_status = is_response_cached(r)
        return UdfEvaluationResult(
            data=data,
            udf=udf,
            time_taken_seconds=time_taken_seconds,
            stdout=meta["stdout"],
            stderr=meta["stderr"],
            error_message=None,
            error_lineno=None,
            has_exception=meta["has_exception"],
            exception_class=meta["exception_class"],
            error_type=meta["error_type"],
            is_cached=cache_status.is_cached,
            cache_source=cache_status.cache_source,
        )

    elif "x-fused-metadata" in r.headers:
        # This is a special case for UDFs that returned an error.
        # We raise a UdfRuntimeError to indicate the runtime error.
        error_message = ""
        if meta["errormsg"]:
            if lineno := meta["error_lineno"]:
                error_message = f"The UDF returned the following error in line {lineno}:\n{meta['errormsg']}"
            else:
                error_message = (
                    f"The UDF returned the following error:\n{meta['errormsg']}"
                )
        elif meta["exception"]:
            error_message = meta["exception"]

        return UdfEvaluationResult(
            data=None,
            time_taken_seconds=time_taken_seconds,
            stdout=meta["stdout"],
            stderr=meta["stderr"],
            error_message=error_message,
            error_lineno=meta["error_lineno"],
            has_exception=meta["has_exception"],
            exception_class=meta["exception_class"],
            error_type=meta["error_type"],
        )

    elif "x-fused-error" in r.headers:
        msg = str(r.headers["x-fused-error"])
        raise requests.HTTPError(msg, response=r)

    raise_for_status(r)


async def _process_response_async(
    r,
    step_config: AnyJobStepConfig,
    time_taken_seconds: float,
) -> UdfEvaluationResult:
    result_content: bytes | None = None
    stdout: str | None = None
    stderr: str | None = None
    error_message: str | None = None
    error_lineno: int | None = None

    try:
        if r.status != 200 and r.status != 422:
            raise ValueError(await r.read())

        result_content = (
            await r.read() if not OPTIONS.pyodide_async_requests else await r.bytes()
        )
        # x-fused-metadata holds LogHandler as JSON, which contains stdout/stderr.
        _fused_metadata = r.headers.get("x-fused-metadata")
        fused_metadata = json.loads(_fused_metadata) if _fused_metadata else {}

        # Extract stdout/stderr.
        stdout = fused_metadata.get("stdout")
        stderr = fused_metadata.get("stderr")

        # Extract exception details
        exception_class = fused_metadata.get("exception_class")
        has_exception = fused_metadata.get("has_exception", False)
        error_type = fused_metadata.get("error_type", None)

        # Extract udf.
        udf: AnyBaseUdf | None = None
        if step_config is not None:
            udf = step_config.udf

        # Extract error line, if exists.
        error_lineno = fused_metadata.get("lineno")

        if r.status == 200:
            # If the UDF returned None.
            if len(result_content) == 0:
                data = None

            # Else, process response output.
            elif r.headers["content-type"] == "image/png":
                data = deserialize_png(result_content)

            elif r.headers["content-type"] == "image/tiff":
                # TODO: Automatically display tiff?
                data = deserialize_tiff(r.content)

            elif r.headers["content-type"] == "application/json":
                data = deserialize_json(result_content)

            elif r.headers["content-type"].startswith("text/html"):
                data = deserialize_html(result_content)

            elif r.headers["content-type"] == "application/zip":
                data = deserialize_zip(result_content)

            else:
                # assume parquet
                data = deserialize_parquet(result_content)

            cache_status = is_response_cached(r)
            return UdfEvaluationResult(
                data=data,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message,
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
                is_cached=cache_status.is_cached,
                cache_source=cache_status.cache_source,
                error_type=error_type,
            )
        else:
            if "errormsg" in fused_metadata and fused_metadata["errormsg"]:
                error_message = f"The UDF returned the following error in line {fused_metadata.get('lineno')}:\n{fused_metadata['errormsg']}"
            elif "exception" in fused_metadata and fused_metadata["exception"]:
                error_message = fused_metadata["exception"]
            else:
                # No error message was returned, e.g. due to deserialization error
                try:
                    # Look for a "detail" field in the response payload
                    details_obj = json.loads(result_content)
                    error_message = details_obj["detail"]
                except:  # noqa: E722
                    error_message = result_content.decode()

            return UdfEvaluationResult(
                data=None,
                udf=udf,
                time_taken_seconds=time_taken_seconds,
                stdout=stdout,
                stderr=stderr,
                error_message=error_message or "Unknown error occured",
                error_lineno=error_lineno,
                has_exception=has_exception,
                exception_class=exception_class,
            )
    except:  # noqa: E722
        r.raise_for_status()
        raise


# Global shared session for async requests
_shared_session: aiohttp.ClientSession | None = None
_session_loop: asyncio.AbstractEventLoop | None = None


async def _get_shared_session() -> aiohttp.ClientSession:
    """Get or create a shared aiohttp session for connection pooling"""
    global _shared_session, _session_loop

    current_loop = asyncio.get_running_loop()

    # Check if we need to create a new session due to:
    # 1. No session exists
    # 2. Session is closed
    # 3. Session was created in a different event loop
    if (
        _shared_session is None
        or _shared_session.closed
        or _session_loop is not current_loop
    ):
        # Clean up old session if it exists
        if _shared_session and not _shared_session.closed:
            await _shared_session.close()

        # Create session with connection pooling settings
        connector = aiohttp.TCPConnector(
            limit=1000,
            ttl_dns_cache=300,
            use_dns_cache=True,
        )
        _shared_session = aiohttp.ClientSession(
            connector=connector,
            timeout=aiohttp.ClientTimeout(total=OPTIONS.request_timeout),
        )
        _session_loop = current_loop

    return _shared_session



from __future__ import annotations

from enum import Enum
from typing import Literal
from urllib.parse import urlparse

from pydantic import BaseModel, Field, StrictBool, StrictStr
from typing_extensions import Annotated

from fused._str_utils import append_url_part


class SampleStrategy(str, Enum):
    """How to generate output samples"""

    EMPTY = "empty"
    """Do not generate a sample"""
    FIRST_CHUNK = "first_chunk"
    """The sample is from the first chunk"""
    GEO = "geo"
    """Geographically sample"""


class DatasetOutputType(str, Enum):
    V2 = "v2"
    """Save as a table to a URL"""


class DatasetOutputBase(BaseModel):
    type: Literal[None, "v2"]

    save_index: StrictBool | None = None
    """Whether to override saving the output index."""

    sample_strategy: SampleStrategy | None = None
    """How to generate output samples, or None for the default."""

    overwrite: bool = False
    """Whether the API should overwrite the output dataset if it already exists."""

    def from_str(s: str | None, project_url: str | None = None) -> AnyDatasetOutput:
        try:
            parsed = urlparse(s)
            if parsed.scheme:
                output = DatasetOutputV2(url=s)
                output._project_url = project_url
                return output
        except (ValueError, TypeError, AttributeError):
            pass

        if project_url is not None:
            url = append_url_part(project_url, s) if s is not None else None
            output = DatasetOutputV2(url=url)
            output._project_url = project_url
            return output

        if s is None:
            return DatasetOutputV2()

        # Parsing as URL failed
        raise ValueError("failed to parse URL")


class DatasetOutputV2(DatasetOutputBase):
    """Output that writes a table to a URL"""

    type: Literal["v2"] = "v2"

    url: StrictStr | None = None
    """Table URL to write to"""

    @property
    def table(self) -> str | None:
        """Returns the table name for this output"""
        if self.url:
            return self.url.rstrip("/").rsplit("/", maxsplit=1)[1]
        return None


AnyDatasetOutput = Annotated[
    DatasetOutputV2,
    Field(..., discriminator="type"),
]



import importlib.abc
import importlib.util
import linecache
import sys
from types import CodeType, ModuleType
from typing import Any, Dict, Literal, Optional

import requests
from pydantic import BaseModel, ConfigDict, PrivateAttr
from typing_extensions import override

from fused._formatter.udf import fused_header_repr
from fused._options import options as OPTIONS
from fused._str_utils import is_url


def _compile_with_line_cache(src: str, module_name: str) -> CodeType:
    file_name = f"<module {module_name}>"
    linecache.cache[file_name] = (
        len(src),
        None,
        [line + "\n" for line in src.splitlines()],
        file_name,
    )
    return compile(src, file_name, mode="exec")


class Header(BaseModel):
    """A header represents a reusable source module included with a UDF."""

    module_name: str
    """The name by which the header may be imported"""
    source_code: Optional[str] = None
    """The code of the header module"""
    source_file: Optional[str] = None
    """The name of the original source file"""
    _finder_instance_id: Optional[int] = PrivateAttr(None)
    """The ID of this header's StringModuleFinder instance, used for cleanup."""
    _prev_sys_module: Optional[ModuleType] = PrivateAttr(None)
    """The module object that was previously in sys.modules for this header's name, if any, saved during registration and restored on exit."""

    def _repr_html_(self) -> str:
        return fused_header_repr([self])

    @classmethod
    def from_code(
        cls, module_name: str, source_file: str, source: Literal["disk", "url"] = "disk"
    ):
        """Read a header from a location.

        Args:
            module_name: The name by which the module may be imported
            source_file: Where to read from
            source: Source type, must be `"disk"` (read a file from disk) or `"url"` (read a file from URL)
        """
        if source == "url" or is_url(source_file):
            return cls._read_code_from_url(module_name, source_file)
        elif source == "disk":
            _header = cls._read_code_from_disk(module_name, source_file)
            if _header:
                # del _header.source_file
                # warnings.warn(
                # "The local header path specified in the UDF source might not resolve if this UDF code is ran elsewhere.\
                # Consider setting a remote header.",
                # FusedWarning,
                # )
                pass
            return _header

    @classmethod
    def _read_code_from_disk(cls, module_name, source_file):
        with open(source_file) as file:
            source_code = file.read()

        return cls(
            module_name=module_name, source_file=source_file, source_code=source_code
        )

    @classmethod
    def _read_code_from_url(cls, module_name, source_file):
        r = requests.get(source_file, timeout=OPTIONS.request_timeout)
        r.raise_for_status()
        source_code = r.text
        return cls(
            module_name=module_name, source_file=source_file, source_code=source_code
        )

    def write_code(self):
        # TODO: select between write_code_to_disk and possible write_code_to_gist
        self._write_code_to_disk()

    def write_to_disk(self, path: str):
        raise NotImplementedError

    def _exec(self) -> Dict[str, Any]:
        """
        Execute the header and return it's locals. This is comparable to importing the header.
        """
        import fused

        ret = {"fused": fused}
        try:
            compiled_code = _compile_with_line_cache(
                src=self.source_code, module_name=self.module_name
            )
            exec(compiled_code, ret)
        except SyntaxError as e:
            # If the syntax error is in the header, make sure it gets a different type so the caller knows
            # it wasn't in the UDF itself.
            raise HeaderSyntaxError(e.msg) from e
        return ret

    def _register_custom_finder(self):
        module_name = self.module_name
        source_code = self.source_code

        class StringModuleFinder(importlib.abc.MetaPathFinder):
            @override
            def find_spec(self, fullname, path, target=None):
                if fullname == module_name:
                    loader = StringModuleLoader(module_name, source_code)
                    spec = importlib.util.spec_from_loader(module_name, loader)
                    return spec
                return None

        # Register the custom finder
        finder_instance = StringModuleFinder()
        self._finder_instance_id = id(finder_instance)
        sys.meta_path.insert(0, finder_instance)

        # Save previous module and reset modules cache
        try:
            self._prev_sys_module = sys.modules.pop(module_name, None)
        except Exception:  # noqa: F841
            pass
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Reset modules
        try:
            sys.meta_path = [
                importer
                for importer in sys.meta_path
                if id(importer) != self._finder_instance_id
            ]
            if self._prev_sys_module:
                sys.modules[self.module_name] = self._prev_sys_module
            else:
                del sys.modules[self.module_name]
        except Exception:  # noqa: F841
            pass

    def _generate_cell_code(self):
        # If remote, comment remote path. Otherwise, set %%file.
        if self.source_file and is_url(self.source_file):
            file_name = self.module_name + ".py"
            return f"%%file {file_name}\n# Remote file: {self.source_file}\n{self.source_code}"
        else:
            file_name = self.source_file or self.module_name + ".py"
            return f"%%file {file_name}\n{self.source_code}"

    @override
    def __getstate__(self):
        state = super().__getstate__()

        new_state = state.copy()
        if "_prev_sys_module" in new_state:
            del new_state["_prev_sys_module"]

        return new_state

    @override
    def __setstate__(
        self,
        state: dict[str, Any],
    ) -> None:
        super().__setstate__(state)

    def __deepcopy__(self, memo=None):
        # Do not use model_copy as it will call __deepcopy__.
        # This will cause issues with _prev_sys_module.
        # TODO: Move this state off the header object itself into a specific context object.
        return Header.model_validate(self.model_dump())

    model_config = ConfigDict(arbitrary_types_allowed=True)


class StringModuleLoader(importlib.abc.Loader):
    def __init__(self, module_name, source_code):
        self.module_name = module_name
        self.source_code = source_code

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        import fused

        module.__dict__["fused"] = fused
        try:
            compiled_code = _compile_with_line_cache(
                src=self.source_code, module_name=self.module_name
            )
            exec(compiled_code, module.__dict__)
        except SyntaxError as e:
            # If the syntax error is in the header, make sure it gets a different type so the caller knows
            # it wasn't in the UDF itself.
            raise HeaderSyntaxError(
                f"{e.msg}: File {e.filename} at line {e.lineno}, position {e.offset}: {e.text.rstrip()}"
            ) from e


class HeaderSyntaxError(Exception):
    pass



from __future__ import annotations

import ast
import json
from pathlib import Path
from typing import Any, Callable, Optional
from zipfile import ZipFile

from loguru import logger

from fused._global_api import get_api
from fused.models._codegen import MetaJson
from fused.models.api import UdfJobStepConfig
from fused.models.udf import (
    AnyBaseUdf,
    Udf,
    load_udf_from_response_data,
)

from .._str_utils import is_uuid
from ._impl._context_impl import context_get_user_email

try:
    from cachetools import TTLCache, cached

    memoize_cache = cached(TTLCache(maxsize=10240, ttl=60))
    logger.debug("cachetools TTL memoize cache initialized")
except ImportError:
    from functools import lru_cache

    # Number of entries to store
    memoize_cache = lru_cache(maxsize=1024)
    logger.debug("lru memoize cache initialized")


def get_step_config_from_server(
    email_or_handle: Optional[str],
    slug: str,
    cache_key: Any,
    _is_public: bool = False,
    import_udf_globals: bool = True,
) -> UdfJobStepConfig:
    logger.info(f"Requesting {email_or_handle=} {slug=}")
    # cache_key is unused
    api = get_api()
    if _is_public:
        obj = api._get_public_udf(slug)
    else:
        obj = api._get_udf(email_or_handle, slug)
    udf = load_udf_from_response_data(
        obj, context={"import_globals": import_udf_globals}
    )

    step_config = UdfJobStepConfig(udf=udf)
    return step_config


def load_udf_from_fused(
    email_or_handle_or_id: str,
    id: str | None = None,
    *,
    cache_key: Any = None,
    import_globals: bool | None = None,
) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_handle_or_id: Email or handle of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    if id is None and not is_uuid(email_or_handle_or_id):
        id = email_or_handle_or_id
        try:
            email_or_handle = context_get_user_email()
        except Exception as e:
            raise ValueError(
                "could not detect user ID from context, please specify the UDF as 'user@example.com' (or 'user'), 'udf_name'."
            ) from e
    else:
        email_or_handle = email_or_handle_or_id
    step_config = get_step_config_from_server(
        email_or_handle=email_or_handle,
        slug=id,
        cache_key=cache_key,
        import_udf_globals=import_globals,
    )

    return step_config.udf


async def get_step_config_from_server_async(
    email_or_handle: Optional[str],
    slug: str,
    cache_key: Any,
    _is_public: bool = False,
    import_udf_globals: bool = True,
) -> UdfJobStepConfig:
    """Async version of get_step_config_from_server that uses async HTTP requests"""
    api = get_api()
    if _is_public:
        obj = api._get_public_udf(slug)
    else:
        obj = await api._get_udf_async(email_or_handle, slug)

    udf = load_udf_from_response_data(
        obj, context={"import_globals": import_udf_globals}
    )
    return UdfJobStepConfig(udf=udf)


async def load_udf_from_fused_async(
    email_or_handle_or_id: str,
    id: str | None = None,
    *,
    cache_key: Any = None,
    import_globals: bool | None = None,
) -> AnyBaseUdf:
    """
    Asynchronously download the code of a UDF, to be run inline.

    This is the async version of load_udf_from_fused that uses async HTTP requests
    to avoid blocking the event loop during UDF metadata fetching.

    Args:
        email_or_handle_or_id: Email or handle of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
        import_globals: Whether to import globals from the UDF context
    """
    if id is None and not is_uuid(email_or_handle_or_id):
        id = email_or_handle_or_id
        try:
            from fused.core._impl._context_impl import context_get_user_email_async

            email_or_handle = await context_get_user_email_async()
        except Exception as e:
            raise ValueError(
                "could not detect user ID from context, please specify the UDF as 'user@example.com' (or 'user'), 'udf_name'."
            ) from e
    else:
        email_or_handle = email_or_handle_or_id

    step_config = await get_step_config_from_server_async(
        email_or_handle=email_or_handle,
        slug=id,
        cache_key=cache_key,
        import_udf_globals=import_globals,
    )
    return step_config.udf


@memoize_cache
def _get_github_udf_from_server(
    url: str, *, cache_key: Any = None, import_globals: bool | None = None
) -> AnyBaseUdf:
    logger.info(f"Requesting {url=}")
    # cache_key is unused
    api = get_api(credentials_needed=False)
    obj = api._get_code_by_url(url)
    udf = load_udf_from_response_data(obj, context={"import_globals": import_globals})
    return udf


def load_udf_from_github(
    url: str, *, cache_key: Any = None, import_globals: bool | None = None
) -> AnyBaseUdf:
    """
    Download the code of a UDF, to be run inline.

    Args:
        email_or_id: Email of the UDF's owner, or name of the UDF to import.
        id: Name of the UDF to import. If only the first argument is provided, the current user's email will be used.

    Keyword args:
        cache_key: Additional cache key for busting the UDF cache
    """
    return _get_github_udf_from_server(
        url=url, cache_key=cache_key, import_globals=import_globals
    )


def load_udf_from_shared_token(
    token: str, import_globals: bool | None = None
) -> AnyBaseUdf:
    """
    Download the code of a UDF from a shared token

    Args:
        token: the shared token for a UDF

    Raises:
        requests.HTTPError if the token is for a UDF that is not owned by the current user
    """
    api = get_api(credentials_needed=False)
    obj = api._get_udf_by_token(token)
    udf = load_udf_from_response_data(obj, context={"import_globals": import_globals})
    return udf


def _get_entrypoint_of_code(code: str):
    """
    Returns the name of the function decorated with '@fused.udf' in the provided code.
    If multiple functions are decorated with '@fused.udf', use "udf" as the entrypoint.
    """
    found_udf_names: list[str] = []
    try:
        tree = ast.parse(code)
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                for decorator in node.decorator_list:
                    is_fused_udf = False
                    actual_decorator = decorator
                    # If the decorator is called (e.g., @fused.udf(arg=...)),
                    # the node is ast.Call, and the actual decorator is in .func
                    if isinstance(decorator, ast.Call):
                        actual_decorator = decorator.func

                    # Check for @fused.udf (Attribute)
                    if isinstance(actual_decorator, ast.Attribute):
                        if (
                            isinstance(actual_decorator.value, ast.Name)
                            and actual_decorator.value.id == "fused"
                            and actual_decorator.attr == "udf"
                        ):
                            is_fused_udf = True
                    # Check for @udf (Name, assuming `from fused import udf`)
                    elif (
                        isinstance(actual_decorator, ast.Name)
                        and actual_decorator.id == "udf"
                    ):
                        is_fused_udf = True

                    if is_fused_udf:
                        found_udf_names.append(node.name)
                        break  # Only need to find one relevant decorator per function

    except SyntaxError as e:
        raise ValueError(f"Invalid Python code provided: {e}") from e

    if not found_udf_names:
        raise ValueError(
            "No function decorated with '@fused.udf' found in the provided code."
        )
    elif len(found_udf_names) == 1:
        return found_udf_names[0]
    elif len(found_udf_names) > 1:
        if "udf" not in found_udf_names:
            raise ValueError(
                f"Multiple functions decorated with '@fused.udf' found: {', '.join(found_udf_names)}. "
                "Please provide code with only one decorated UDF."
            )
        else:
            return "udf"


def load_udf_from_code(
    code: str, name: Optional[str] = None, import_globals: bool | None = None
) -> Udf:
    """
    Load a UDF from raw code.

    """
    udf_entrypoint_name = _get_entrypoint_of_code(code)

    data = {
        "name": name or udf_entrypoint_name,  # Use found name as default
        "entrypoint": udf_entrypoint_name,  # Use found name as entrypoint
        "type": "geopandas_v2",
        "code": code,
    }
    return Udf.model_validate(
        data, context={"import_globals": import_globals, "load_parameter_list": True}
    )


def load_udf_from_file(path: Path, import_globals: bool | None = None) -> Udf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """
    code = path.read_bytes().decode("utf8")
    udf_entrypoint_name = _get_entrypoint_of_code(code)

    data = {
        "name": path.stem,
        "entrypoint": udf_entrypoint_name,
        "type": "geopandas_v2",
        "code": code,
    }
    return Udf.model_validate(
        data, context={"import_globals": import_globals, "load_parameter_list": True}
    )


def _get_udf_from_directory(
    load_callback: Callable[[str], bytes], import_globals: bool | None = None
) -> Udf:
    meta_contents = json.loads(load_callback("meta.json"))
    meta = MetaJson.model_validate(meta_contents)

    if len(meta.job_config.steps) != 1:
        raise ValueError(
            f"meta.json is not in expected format: {len(meta.job_config.steps)=}"
        )

    if meta.job_config.steps[0]["type"] != "udf":
        raise ValueError(
            f'meta.json is not in expected format: {meta.job_config.steps[0]["type"]=}'
        )

    # Load the source code into the UDF model
    udf_dict = meta.job_config.steps[0]["udf"]
    source_file_name = udf_dict["source"]

    code = load_callback(source_file_name).decode("utf-8")
    udf_dict["code"] = code
    del udf_dict["source"]

    # Do the same for headers
    for header_dict in udf_dict["headers"]:
        header_source_file_name = header_dict.get("source_file")
        if header_source_file_name:
            del header_dict["source_file"]
            header_code = load_callback(header_source_file_name).decode("utf-8")
            header_dict["source_code"] = header_code

    return Udf.model_validate(
        udf_dict,
        context={"import_globals": import_globals, "load_parameter_list": True},
    )


def load_udf_from_directory(
    path: Path, import_globals: bool | None = None
) -> AnyBaseUdf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """

    def _load_file(name: str) -> bytes:
        file_path = path / name
        if not file_path.exists():
            raise ValueError(
                f"Expected a file to be at {repr(file_path)}. Is this the right directory to load from?"
            )

        return file_path.read_bytes()

    return _get_udf_from_directory(
        load_callback=_load_file, import_globals=import_globals
    )


def load_udf_from_zip(path: Path, import_globals: bool | None = None) -> AnyBaseUdf:
    """
    Load a UDF from a python file

    Args:
        path : pathlib.Path
    """
    with ZipFile(path) as zf:
        return _get_udf_from_directory(
            load_callback=lambda f: zf.read(f), import_globals=import_globals
        )



"""Vendored from xarray
https://github.com/pydata/xarray/blob/f4e0523be0ce8babaa8eff38365e5308b1fdb76b/xarray/core/formatting_html.py
"""

from __future__ import annotations

import uuid
from html import escape
from typing import TYPE_CHECKING

from fused._formatter.common import copyable_text, icon, load_static_files

if TYPE_CHECKING:
    from fused.models.api.udf_access_token import UdfAccessToken, UdfAccessTokenList


def _obj_repr(obj, header_components, sections):
    """Return HTML repr of an xarray object.

    If CSS is not injected (untrusted notebook), fallback to the plain text repr.

    """
    header = f"<div class='xr-header'>{''.join(h for h in header_components)}</div>"
    sections = "".join(f"<li class='xr-section-item'>{s}</li>" for s in sections)

    icons_svg, css_style = load_static_files()
    return (
        "<div>"
        f"{icons_svg}<style>{css_style}</style>"
        f"<pre class='xr-text-repr-fallback'>{escape(repr(obj))}</pre>"
        "<div class='xr-wrap' style='display:none'>"
        f"{header}"
        f"<ul class='xr-sections'>{sections}</ul>"
        "</div>"
        "</div>"
    )


def summarize_udf_access_token(
    token: UdfAccessToken,
    *,
    is_index: bool = False,
    dtype=None,
):
    cssclass_idx = " class='xr-has-index'" if is_index else ""
    dims_str = ""
    name = escape(token.token)
    dtype = escape(f"{token.udf_email}/{token.udf_slug}")
    preview = escape(token.client_id) if token.client_id else ""

    # "unique" ids required to expand/collapse subsections
    attrs_id = "attrs-" + str(uuid.uuid4())
    data_id = "data-" + str(uuid.uuid4())

    configuration: list[str] = []
    for key in token.model_dump().keys():
        value = getattr(token, key)
        if key == "last_updated":
            configuration.append(
                f"<li>{key}: <code>{escape(value.isoformat())}</code></li>"
            )
        elif key not in ["token", "udf_email", "udf_slug"]:
            configuration.append(f"<li>{key}: <code>{escape(repr(value))}</code></li>")

    # TODO: Fix for no schema, fix for parameters expected by the UDF, fix for copy decorator, fix for entrypoint
    attrs_ul = f"""
    <h4 style="margin-left: 20px;">Token</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    <li>{copyable_text(token.token, show_text=True)}</li>
    <li>UDF owner email: <code>{escape(token.udf_email) if token.udf_email else "None"}</code></li>
    <li>UDF name: <code>{escape(token.udf_slug) if token.udf_slug else "Unknown"}</code></li>
    <li>UDF ID: <code>{escape(token.udf_id) if token.udf_id else "Unknown"}</code></li>
    </ul>
    <h4 style="margin-left: 20px;">Configuration</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    {"".join(configuration)}
    </ul>
    <h4 style="margin-left: 20px;">Toolkit</h4>
    <ul style="margin-top: 0; margin-bottom: 0px;">
    <li>File URL: {copyable_text(token.get_file_url())}</li>
    <li>Tile URL: {copyable_text(token.get_tile_url())}</li>
    </ul>
    """
    attrs_icon = icon("icon-file-text2")
    code_icon = icon("icon-code")

    chunk_repr_disabled = "disabled"
    data_repr = ""

    return (
        f"<div class='xr-var-name'><span{cssclass_idx}>{name}</span></div>"
        f"<div class='xr-var-dims'>{dims_str}</div>"
        f"<div class='xr-var-dtype'>{dtype}</div>"
        f"<div class='xr-var-preview xr-preview'>{preview}</div>"
        f"<input id='{attrs_id}' class='xr-var-attrs-in' "
        f"type='checkbox'>"
        f"<label for='{attrs_id}' title='Show/Hide attributes'>"
        f"{attrs_icon}</label>"
        f"<input id='{data_id}' class='xr-var-data-in' type='checkbox' {chunk_repr_disabled}>"
        f"<label for='{data_id}' title='Show/Hide data repr' {chunk_repr_disabled}>"
        f"{code_icon}</label>"
        f"<div class='xr-var-attrs'>{attrs_ul}</div>"
        f"<div class='xr-var-data'>{data_repr}</div>"
    )


def _udf_access_token_repr_dynamic(obj: UdfAccessToken) -> str:
    ret: list[str] = []

    for key in obj.model_dump().keys():
        value = getattr(obj, key)
        if key == "last_updated":
            ret.append(f"{key}: <code>{escape(value.isoformat())}</code>")
        elif key not in ["token", "udf_email", "udf_slug"]:
            ret.append(f"{key}: <code>{escape(repr(value))}</code>")
    return f'<div style="grid-column: 1 / span 4;"><ul>{"".join([f"<li>{r}</li>" for r in ret])}</ul></div>'


def _udf_access_token_repr_static(obj: UdfAccessToken) -> str:
    ret: list[str] = []

    ret = [
        f"UDF Email: <code>{escape(obj.udf_email)}</code>",
        f"UDF ID: <code>{escape(obj.udf_slug)}</code>",
    ]

    return f'<div style="grid-column: 1 / span 4;"><ul>{"".join([f"<li>{r}</li>" for r in ret])}</ul></div>'


def fused_udf_access_token_repr(token: UdfAccessToken) -> str:
    header_components = [
        f"<div class='xr-obj-type'>UdfAccessToken: {escape(f'{token.udf_email}/{token.udf_slug}')}</div>"
    ]

    sections = [summarize_udf_access_token(token)]
    # sections.append(_udf_access_token_repr_static(token))
    # sections.append(_udf_access_token_repr_dynamic(token))

    return _obj_repr(token, header_components, sections)


def fused_udf_access_token_list_repr(tokens: UdfAccessTokenList) -> str:
    header_components = [
        f"<div class='xr-obj-type'>UdfAccessTokens: {len(tokens)}</div>"
    ]

    sections: list[str] = []
    for token in tokens:
        sections.append(summarize_udf_access_token(token))

    return _obj_repr(tokens, header_components, sections)
